import { wrapperLoggedInUser } from "lib/route_wrapper/wrapperLoggedIn";
import { RequestProps } from "lib/route_wrapper/wrapperSession";
import { NextRequest, NextResponse } from "next/server";
import { z } from "zod";

import service from "@snip/database";
import { FONT_URLS } from "@snip/database/fonts/internal";
import { pool_permissions as pool } from "@snip/database/sql.connection";
import { UserConfigDataRet } from "@snip/database/types";

export const GET = wrapperLoggedInUser(GetUserConfig);

export const POST = wrapperLoggedInUser(EditUserConfig);

async function GetUserConfig(
    _req: NextRequest,
    { session }: DeepPartialExcept<RequestProps, "session.user">,
) {
    let config: UserConfigDataRet | undefined;
    try {
        config = await service.userConfig.getByUserId(session.user.id);
    } catch (e) {
        if (e instanceof Error && e.message.includes("not found")) {
            // Create default config
            config = await service.userConfig.insert({
                user_id: session.user.id,
            });
        } else {
            return NextResponse.json(
                {
                    error: "Internal Server Error",
                    details: "Error fetching user config",
                },
                { status: 500 },
            );
        }
    }
    return NextResponse.json<UserConfigDataRet>(config, { status: 200 });
}

const colorRegex = /^#[0-9A-F]{6}$/i;

const EditUserConfigSchema = z.object({
    zoom1: z.number().optional(),
    zoom2: z.number().optional(),
    text_lineWrap: z.number().optional(),
    text_lineHeight: z.number().optional(),
    text_fontSize: z.number().optional(),
    text_fontColor: z.string().regex(colorRegex).optional(),
    text_font: z.string().optional(),
    text_font_id: z.number().optional(),
    pen_size: z.number().optional(),
    pen_color: z.string().regex(colorRegex).optional(),
    pen_colors: z.array(z.string().regex(colorRegex)).max(128).optional(),
    pen_smoothing: z.number().min(0).max(1).optional(),
});

async function EditUserConfig(
    req: NextRequest,
    { session }: DeepPartialExcept<RequestProps, "session.user">,
) {
    // Check if any updates are given
    const update_config = (await req.json()) as unknown;

    // Validate
    const result = EditUserConfigSchema.safeParse(update_config);

    if (result.success === false) {
        return NextResponse.json(
            {
                error: "Bad Request",
                details: result.error.format(),
            },
            { status: 400 },
        );
    }

    const config = result.data;

    if (Object.keys(config).length === 0) {
        return NextResponse.json(
            {
                error: "Bad Request",
                details: "No updates given",
            },
            { status: 400 },
        );
    }

    // font to font id
    // Get id and delete font name
    if (config.text_font) {
        const found = FONT_URLS.find((font) => font[1] == config.text_font);

        if (!found) {
            return NextResponse.json(
                {
                    error: "Bad Request",
                    details:
                        "text_font must be a valid font, found: " +
                        config.text_font +
                        " (not in (" +
                        FONT_URLS.map((font) => font[1]).join(", ") +
                        "))",
                },
                { status: 400 },
            );
        }
        try {
            await pool
                .query("SELECT id FROM fonts WHERE name = ?", [
                    config.text_font,
                ])
                .then((result) => {
                    config.text_font_id = result[0].id;
                    delete config["text_font"];
                });
        } catch (_e) {
            return NextResponse.json(
                {
                    error: "Internal Server Error",
                    details: "Error fetching font",
                },
                { status: 500 },
            );
        }
    }

    // Update
    const user_config = await service.userConfig.updateByUserId({
        ...result.data,
        user_id: session.user.id,
    });

    return NextResponse.json<UserConfigDataRet>(user_config, { status: 200 });
}
